#include<cstdio>
#include<cmath>
#include<vector>
#include<algorithm>
using namespace std;
#define in stdin;
#define pb push_back
#define sz size()
#define eps 1e-8

FILE*fn;
struct point
{
double x,y;
point(){x=y=0.0;};
point(double a,double b){x=a;y=b;};
};
bool eq(double a,double b){if(a-b<eps&&b-a<eps)return 1;return 0;}

point operator^(const point&a,const double&b){return point(cos(b)*a.x-sin(b)*a.y,sin(b)*a.x+cos(b)*a.y);}
point operator+(const point&a,const point&b){return point(a.x+b.x,a.y+b.y);}
point operator-(const point&a,const point&b){return point(a.x-b.x,a.y-b.y);}
point operator*(double&a,const point&b){return point(a*b.x,a*b.y);}
double operator!(const point&a){return hypot(a.x,a.y);}
double operator^(const point&a,const point&b){return a.x*b.y-a.y*b.x;}
double operator~(const point&a){return atan2(a.y,a.x);}
bool operator==(const point&a,const point&b){return eq(a.x,b.x)&&eq(a.y,b.y);}
bool operator>(const point&a,const point&b){return (a.y>b.y&&eq(a.y,b.y)==0)||(eq(a.y,b.y)&&a.x>b.x);}
bool operator<(const point&a,const point&b)
  {return ((~a<~b)&&(!eq(~a,~b)))||(eq(~a,~b)&&!a<!b);}

int cmp(const void*aa,const void*bb)
  {point a=*((point*)aa),b=*((point*)bb);
  if(( ((~a)<(~b)) &&(!eq(~a,~b)))||(eq(~a,~b)&&((!a)<(!b))))return -1;
  return 1;}

double det2(point&a,point&b,point&c)
  {return (b-a)^(c-a);}


struct points
{
vector < point > ss;
vector < int > hull;
vector < double > hang;
vector < double > dist;
vector < double > ang;
vector < int > km;

int sortit(void)
  {int q1,c,c1;
  c=ss.sz;
  for(c1=0,q1=1;q1<c;q1++)if(ss[c1]>ss[q1])c1=q1;
  swap(ss[0],ss[c1]);
  for(q1=1;q1<=c;q1++)ss[q1%c]=ss[q1%c]-ss[0];
//  qsort(&ss[1],c-1,sizeof(ss[0]),cmp);
  sort(ss.begin()+1,ss.end());
  return 0;}
  
int readit(void)
  {int q1,c;
  ss.clear();
  hull.clear();
  hang.clear();
  dist.clear();
  ang.clear();
  km.clear();
  fscanf(fn,"%d",&c);
  for(q1=0;q1<c;q1++)
    {point a;
    fscanf(fn,"%lf %lf",&a.x,&a.y);
    ss.pb(a);}
//  printf("here1\n");
  return 0;}

int reflect(void)
  {int q1,c;
  hull.clear();
  hang.clear();
  dist.clear();
  ang.clear();
  km.clear();
  c=ss.sz;
  for(q1=0;q1<c;q1++)ss[q1].x=-ss[q1].x;
  sortit();
  return 0;}

int findconvex(void)
  {int q1,c=ss.sz,hz;
  hull.pb(0);
  hull.pb(1);
  hull.pb(2);
  hz=3;
//  printf("here\n");
  for(q1=3;q1<c;q1++)
    {while(hz>=2)
      {double c1=det2(ss[q1],ss[hull[hz-1]],ss[hull[hz-2]]);
      if(c1>0.0||eq(c1,0.0))
        {hz--;
        hull.pop_back();
        continue;}
      break;}
    hull.pb(q1);
    hz++;}

  c=hull.sz;
  hull.pb(hull[0]);hull.pb(hull[1]);
  
  hang.resize(c);
  
  for(q1=1;q1<=c;q1++)
    {double c1,c2,c3,c4;
    c1=!(ss[hull[q1+1]]-ss[hull[q1-1]]);
    c2=!(ss[hull[q1]]-ss[hull[q1+1]]);
    c3=!(ss[hull[q1-1]]-ss[hull[q1]]);
    c4=(c2*c2+c3*c3-c1*c1)/2.0/c2/c3;
    hang[q1%c]=acos(c4);}

  hull.pop_back();
  hull.pop_back();
  return 0;}

int findang(void)
  {int q1,c=ss.sz;

  ang.pb(0.0);
  dist.pb(0.0);
  for(q1=1;q1<c;q1++)
    {ang.pb(-atan2(ss[1].y,ss[1].x)+atan2(ss[q1].y,ss[q1].x));
    dist.pb(!ss[q1]);}
    
  return 0;}

int findkmp(void)
  {int q1,q2,c1,c2,c=hang.sz;
  km.resize(hang.sz);

  km[0]=q2=-1;
  for(q1=1;q1<c;q1++)
    {while(q2!=-1&&eq(hang[q1],hang[q2+1])==0)q2=km[q2];
    if(eq(hang[q1],hang[q2+1]))q2++;
    km[q1]=q2;}
  return 0;}

int printit(void)
  {int q1,c;
  printf("\n");
  c=ss.sz;for(q1=0;q1<c;q1++)printf("%.3lf %.3lf   ",this->ss[q1].x,this->ss[q1].y);printf("\n");
  c=hull.sz;for(q1=0;q1<c;q1++)printf("%d ",this->hull[q1]);printf("\n");
  c=hang.sz;for(q1=0;q1<c;q1++)printf("%.4lf ",this->hang[q1]);printf("\n");
  c=ang.sz;for(q1=0;q1<c;q1++)printf("%.4lf ",this->ang[q1]);printf("\n");
  c=dist.sz;for(q1=0;q1<c;q1++)printf("%.4lf ",this->dist[q1]);printf("\n");
  
  return 0;}

};

points rotate(const points&a,int f,int t)
  {points b;
  int q1,c=a.ss.sz;
  b.ss=a.ss;
  for(q1=0;q1<c;q1++)b.ss[q1]=b.ss[q1]-a.ss[f];
  double c1,c2;
  c1=atan2(b.ss[t].y,b.ss[t].x);
 // c1-=M_PI/2.0;
  for(q1=0;q1<c;q1++)b.ss[q1]=b.ss[q1]^(-c1);

  b.sortit();
  b.findconvex();
  b.findang();
  
  return b;}

vector < int > where(points&a,points&b)
  {int q1,q2,c1,c2,c=a.hang.sz;
  vector < int > ans;
  if(a.hang.sz!=b.hang.sz)return ans;
  q2=-1;
  for(q1=0;q1<2*c;q1++)
    {while(q2!=-1&&eq(b.hang[q2+1],a.hang[q1%c])==0)q2=b.km[q2];
    if(eq(b.hang[q2+1],a.hang[q1%c]))q2++;
    if(q2==b.hang.sz-1)
      ans.pb((2*c-q1-1)%c);
    }
  
  return ans;}

int check(const points&a,const points&b)
  {int q1,c=a.ang.sz;
  double ratio=a.dist[1]/b.dist[1];
  for(q1=1;q1<c;q1++)
    {if(!eq(a.ang[q1],b.ang[q1]))return 0;
    if(!eq(ratio,a.dist[q1]/b.dist[q1]))return 0;}

  return 1;}

points ss,s,s1;


int c,q1,q2,ok,times;
int main()
  {fn=in;
//  printf("fdsfdsfds\n");
  ss.readit();
  ss.sortit();
  ss.findconvex();
  ss.findang();
//  ss.printit();
  
  fscanf(fn,"%d",&c);
  for(q1=0;q1<c;q1++)
    {ok=0;
    s.readit();
    if(s.ss.sz!=ss.ss.sz)
      {printf("NIE\n");
      continue;}
    if(s.ss.sz==1&&ss.ss.sz==1)
      {printf("TAK\n");
      continue;}
    s.reflect();
    s.sortit();
    s.findconvex();
    s.findang();
   // s.printit();
    s.findkmp();
    vector < int > k;
    k=where(ss,s);
    for(q2=0;q2<k.sz&&ok==0;q2++)
      {s1=rotate(s,s.hull[k[q2]],s.hull[(k[q2]+1)%s.hull.sz]);
      //s1.printit();
      times++;
      if(check(ss,s1)){printf("TAK\n");ok=1;break;}}

    if(ok)continue;

    s.reflect();
    s.sortit();
    s.findconvex();
    s.findang();
   // s.printit();
    s.findkmp();
    k=where(ss,s);
    for(q2=0;q2<k.sz&&ok==0;q2++)
      {s1=rotate(s,s.hull[k[q2]],s.hull[(k[q2]+1)%s.hull.sz]);
      //s1.printit();
      times++;
      if(check(ss,s1)&&ok==0){printf("TAK\n");ok=1;break;}}
    if(ok==0)
      printf("NIE\n");
    }
  fclose(fn);
  printf("%d\n",times);
  return 0;}
