(define (cube a) (* a a a))
(define (square a ) (* a a))
(define (inc a) (+ a 1))
(define (id x ) x)
(define eps 0.0000001)

(define (sum-integers a b)
  (if (> a b)
      0
      (+ a (sum-integers (+ a 1) b))))

(define (sum-cubes a b)
  (if (> a b)
      0
      (+ (cube a) (sum-cubes (+ a 1) b))))

(define (pi-sum a b)
  (if (> a b)
      0
      (+ (/ 1.0 (* a (+ a 2))) (pi-sum (+ a 4) b))))

(define (sum term a next b)
  (if (> a b)
      0
      (+ (term a)
         (sum term (next a) next b))))

(define (sum-cubes2 a b)
  (sum cube a inc b))

(define (sum-integers2 a b)
  (sum id a inc b))

(define (pi-sum2 a b)
  (define (pi-term a) 
    (/ 1.0 (* a (+ a 2))))
  (define (pi-next a) (+ 4 a))
  (sum pi-term a pi-next b))


(define (integral f a b dx)
  (define (adddx a) (+ a dx))
  (* (sum f (+ a (/ dx 2.0)) adddx b)
     dx))

(define (simpson f a b n)
  (define (gety x k)
    (* (f x)
       (cond ((or (= k 0) (= k n)) 1.0)
             ((= (remainder k 2) 1) 4.0)
             ((= (remainder k 2) 0) 2.0)
             (else 3.0))))
  (define (calc k)
    (if (> k n)
        0.0
        (+ (gety (+ a (* k (/ (+ a b) n))) k)
           (calc (+ 1 k)))))
  (* (/ (/ (+ a b) n) 3) (calc 0)))
       

(define (sum2 term a next b)
  (define (iter a result)
    (if (> a b)
        result
        (iter (next a) (+ result (term a)))))
  (iter a 0))

(define (product term a next b)
  (define (iter a result)
    (if (> a b)
        result
        (iter (next a) (* result (term a)))))
  (iter a 1))

(define (product2 term a next b)
  (define (iter a result)
    (if (> a b)
        result
        (iter (next a) (* result (term a)))))
  (iter a 1))

(define (calc-pi k)
  (define (term-pi x) (/ (* x (+ 2.0 x)) (square (+ 1.0 x))))
  (* 4.0
     (product2 term-pi 2.0 (lambda (x) (+ 2.0 x)) k)))

(define (accu combiner start term a next b)
  (if (> a b)
      start
      (combiner (term a)
         (accu combiner start term (next a) next b))))

(define (accu2 combiner start term a next b)
  (define (iter a result)
    (if (> a b)
        result
        (iter (next a) (combiner result (term a)))))
  (iter a start))

(define adder (lambda (x y) (+ x y)))
(define multiplier (lambda (x y) (* x y)))
(define (id x ) x)

(define (prime? a)
  (define (iter b)
    (if (> (* b b) a)
        #t
        (and (not (= (remainder a b) 0)) (iter (+ b 1)))))
  (cond ((= a 1) #f)
        (else (iter 2))))

(define (gcd a b)
  (if (= b 0)
      a
      (gcd b (remainder a b))))

(define (filtered-accumulate filter combiner start term a next b)
  (if (> a b)
      start
      (combiner (if (filter a) (term a) start)
         (filtered-accumulate filter combiner start term (next a) next b))))

(define (sumofprimes a b)
  (filtered-accumulate prime? adder 0 square a inc b))

(define (prodofgcds n)
  (filtered-accumulate (lambda (x) (= (gcd x n) 1)) multiplier 1 id 1 inc n))

(define (pi-sum3 a b)
  (sum2 (lambda (x) (/ 1.0 (* x (+ x 2))))
        a
        (lambda (x) (+ x 4))
        b))

(define (integral2 f a b dx)
  (* (sum f
          (+ a (/ dx 2.0))
          (lambda (x) (+ x dx))
          b)
     dx))

(display "computing f(x,y)=x(1+xy)^2+y(1-y)+(1+xy)(1-y)\n")

(define (f1 x y)
  (define (doit a b)
    (+ (* x a a)
       (* y b)
       (* a b)))
  (doit (+ 1 (* x y)) (- 1 y)))

(define (f2 x y)
  ((lambda (a b)
     (+ (* x a a)
        (* y b)
        (* a b)))
   (+ 1 (* x y))
   (- 1 y)))

(define (f3 x y)
  (let ((a (+ 1 (* x y)))
        (b (- 1 y)))
    (+ (* x a a)
       (* y b)
       (* a b))))

(define (f4 x y)
  (define a (+ 1 (* x y )))
  (define b (- 1 y))
  (+ (* x a a)
       (* y b)
       (* a b)))

(define (close x y)
  (< (abs ( - x y)) eps))

(define (search f a b)
  (let ((mid (/ (+ a b) 2.0)))
    (if (close a b)
        mid
        (let ((test (f mid)))
          (cond ((> test 0.0) 
                 (search f a mid))
                ((< test 0.0)
                 (search f mid b))
                (else mid))))))

(define (half f a b)
  (let ((aa (f a))
        (bb (f b)))
    (cond ((and (<= aa 0.0) (>= bb 0.0)) 
           (search f a b))
          ((and (>= aa 0.0) (<= bb 0.0)) 
           (search f b a))
          (else 
           (display "Try another values for a and b\n")))))

(define (fixed f a)
  (define (try b)
    (let ((next (f b)))
      (if (close next b)
          next
          (try next))))
  (try a))

(define (sqrt2 x)
  (fixed (lambda (y) (/ (+ y (/ x y)) 2.0))
         1.0))

(define (fixedp f a)
  (define (try b)
    (display b)
    (display "\n")
    (let ((next (f b)))
      (if (close next b)
          next
          (try next))))
  (try a))

(define (soltoxx1000damp) (fixedp (lambda (x) (/ (+ (/ (log 1000) (log x)) x) 2.0)) 1.1))

(define (cont-frac n d k)
  (define (iter kk)
    (if (> kk k)
        0.0
        (/ (n kk)
           (+ (d kk)
              (iter (+ 1 kk))))))
  (iter 1))

(define (cont-frac2 n d k)
  (define (iter kk res)
    (if (= kk 0)
        res
        (iter (- kk 1)
              (/ (n kk)
                 (+ (d kk) 
                    res)))))
  (iter k 0.0))

(define (oneoverphi) (cont-frac (lambda (x) 1.0) (lambda (x) 1.0) 10000))

(define (calc-e a)
  (define (dd x)
    (cond ((= (remainder x 3) 1) 1.0)
          ((= (remainder x 3) 0) 1.0)
          (else (* (/ (+ 1.0 x) 3.0) 2.0))))
  (define (nn x) 1.0)
  (+ 2.0
     (cont-frac2 nn dd a)))

(define (cont-frac-minus n d k)
  (define (iter kk res)
    (if (= kk 0)
        res
        (iter (- kk 1)
              (/ (n kk)
                 (- (d kk) 
                    res)))))
  (iter k 0.0))

(define (calc-tan x k)
  (define (nn kk)
    (if (= kk 1)
        x
        (* x x)))
  (define (dd kk)
    (- (* 2 kk) 1))
  (cont-frac-minus nn dd k))

(define (avgdamp f)
  (lambda (x) (/ (+ x (f x)) 2.0)))

(define (sqrt4 x)
  (fixed (avgdamp (lambda (y) (/ x y)))
         1.0))

(define (cbrt4 x)
  (fixed (avgdamp (lambda (y) (/ x (square y))))
         1.0))

(define (deriv a)
  (lambda (x) (/ (- (a (+ eps x)) (a x)) eps)))

(define (newtontrans g)
  (lambda (x)
    (- x
       (/ (g x)
          ((deriv g) x)))))

(define (newton f g)
  (fixed (newtontrans f) g))

(define (sqrt5 x)
  (newton (lambda (y) (- (square y) x)) 1.0))

(define (cubic a b c) 
  (lambda (x)
    (+ (* 1 x x x)
       (* a x x)
       (* b x)
       (* c))))

(define (double f)
  (lambda (x) 
    (f (f x))))

(define (compose f g)
  (lambda (x) (f (g x))))

(define (composen f k)
  (define (iter kk)
    (if (> kk k)
        (lambda (x) x)
        (compose f (iter (+ 1 kk)))))
  (iter 1))

(define (smooth a)
  (lambda (x)
    (/ (+ (a (- x eps)) 
          (a (+ x eps))
          (a (- x   0)))
       3.0)))

(define (smoothn a n)
  (define (iter kk)
    (if (> kk n)
        a
        (smooth (iter (+ kk 1)))))
  (iter 1))



;iterative-improve
(define (iterative-improve good improve)
  (lambda (n)
    ((lambda (a) (a a n))
     (lambda (x n)
       (if (good n) 
           n
           (x x (improve n)))))))

(define (sqrt5 x)
  (define (improvesqrt guess)
    (/ (+ guess (/ x guess)) 2.0))
  (define (goodsqrt guess)
    (close (square guess) x))
  ((iterative-improve goodsqrt improvesqrt) x))

(define (fixed3 f a)
  (define (improvefixed x)
    (f x))
  (define (goodfixed x)
    (close x (f x)))
  ((iterative-improve goodfixed improvefixed) a))

(sqrt5 10)
;EXTRA FOR EXPERTS
;recursion without define :) (only lambdas)
(define (fac n)
  (let (( a 
          (lambda (x n)
            (if (= n 0) 
                1
                (* n (x x (- n 1)))))))
    (a a n)))

(define (fac2 n)
  ((lambda (a) (a a n))
   (lambda (x n)
     (if (= n 0) 
         1
         (* n (x x (- n 1)))))))

(define fac3
  (lambda (n)
    ((lambda (a) (a a n))
     (lambda (x n)
       (if (= n 0) 
           1
           (* n (x x (- n 1))))))))

; 1.45 / p.78

(define (dampn f n)
  ((composen avgdamp n) f))

(define (expo x n)
  (if (= n 0)
      1
      (* x (expo x (- n 1 )))))

(define (exp n)
  (lambda (x) (expo x n)))

(define (fixedexp x m damps)
  (dampn (lambda (y) (/ x (expo y (- m 1)))) damps))

(define (roott x n damps)
  (fixed (fixedexp x n damps) 1.0))

(define (log2 a)
  (if (< a 0.9)
      -1
      (+ 1 (log2 (/ a 2)))))

(define (root x n)
  (roott x n (log2 n)))