(define (equal l1 l2)
  (cond ((and (pair? (car l1)) (pair? (carl2))) (and (equal (car l1) (car l2)) (equal (cdr l1) (cdr l2))))
        ((and (not (pair? (car l1))) (pair? (car l2))) #f)
        ((and (not (pair? (car l2))) (pair? (car l1))) #f)
        (else (and (eq? (car l1) (car l2)) (equal? (cdr l1) (cdr l2))))))

(define (deriv exp var)
  (cond ((number? exp) 0)
        ((variable? exp)
         (if (same-variable? exp var) 1 0))
        ((sum? exp)
         (make-sum (deriv (addend exp) var)
                   (deriv (augend exp) var)))
        ((product? exp)
         (make-sum
          (make-product (multiplier exp)
                        (deriv (multiplicand exp) var))
          (make-product (deriv (multiplier exp) var)
                        (multiplicand exp))))
        ((exponentiation? exp)
         (make-product (make-product (exponent exp)
                                     (make-exponentiation (base exp) (- (exponent exp) 1)))
                       (deriv (base exp) 'x)))
                                                                
        (else "i do not know")))

(define (variable? x) (symbol? x))
(define (same-variable? v1 v2)
  (and (variable? v1) (variable? v2) (eq? v1 v2)))
(define (make-sum c1 c2) (list '+ c1 c2))
(define (make-product c1 c2) (list '* c1 c2))
(define (make-exponentiation c1 c2) (list '** c1 c2))
(define (sum? x)
  (and (pair? x) (eq? (car x) '+)))
(define (product? x) 
  (and (pair? x) (eq? (car x) '*)))
(define (exponentiation? x)
  (and (pair? x) (eq? (car x) '**) (= (length x) 3) (number? (caddr x))))
(define (addend c1) (cadr c1))
(define (augend c1) 
  (cond ((null? (cddr c1)) 0) 
        ((null? (cdddr c1)) (caddr c1))
        (else (cons (car c1) (cddr c1)))))
(define (multiplier c1) (cadr c1))
(define (multiplicand c1) 
  (cond ((null? (cddr c1)) 1) 
        ((null? (cdddr c1)) (caddr c1))
        (else (cons (car c1) (cddr c1)))))
(define (base c1) (cadr c1))
(define (exponent c1) (caddr c1))

(deriv '(* (* x y) (+ 3 x)) 'x)

(define (make-sum c1 c2)
  (cond ((eq? c1 0) c2)
        ((eq? c2 0) c1)
        ((and (number? c1) (number? c2)) (+ c1 c2))
        ((and (sum? c1) (sum? c2)) (append c1 (cdr c2)))
        ((sum? c1) (append c1 (list c2)))
        ((sum? c2) (append (list '+ c1) (cdr c2)))
        (else (list '+ c1 c2))))
 
(deriv '(* (* x y) (+ 3 x)) 'x)

(define (make-product c1 c2)
  (cond ((eq? c1 0) 0)
        ((eq? c2 0) 0)
        ((eq? c1 1) c2)
        ((eq? c2 1) c1)
        ((and (number? c1) (number? c2)) (* c1 c2))
        ((and (product? c1) (product? c2)) (append c1 (cdr c2)))
        ((product? c1) (append c1 (list c2)))
        ((product? c2) (append (list '* c1) (cdr c2)))
        (else (list '* c1 c2))))

(define (make-exponentiation c1 c2) 
  (cond ((eq? c2 1) c1)
        ((eq? c2 0) 1)
        (else (list '** c1 c2))))

(deriv '(* (* x y) (+ 3 x)) 'x)

(define (insum? x)
  (and (list? x) (memq '+ x)))
(define (inproduct? x)
  (and (list? x) (memq '* x) (not (insum? x))))
(define (inaddendh x)
  (if (eq? (car x) '+)
      null
      (cons (car x) (inaddendh (cdr x)))))
(define (inaddend x)
  (let ((a (inaddendh x)))
    (if (null? (cdr a)) (car a) a)))
(define (inaugendh x)
  (cdr (memq '+ x)))
(define (inaugend x)
  (let ((a (inaugendh x)))
    (if (null? (cdr a)) (car a) a)))
(define (inmultiplierh x)
  (if (eq? (car x) '*)
      null
      (cons (car x) (inmultiplierh (cdr x)))))
(define (inmultiplier x)
  (let ((a (inmultiplierh x)))
    (if (null? (cdr a)) (car a) a)))
(define (inmultiplicandh x)
  (cdr (memq '* x)))
(define (inmultiplicand x)
  (let ((a (inmultiplicandh x)))
    (if (null? (cdr a)) (car a) a)))

(define (inmake-product c1 c2)
  (cond ((eq? c1 0) 0)
        ((eq? c2 0) 0)
        ((eq? c1 1) c2)
        ((eq? c2 1) c1)
        ((and (number? c1) (number? c2)) (* c1 c2))
        ((and (product? c1) (inproduct? c2)) (append c1 (cons '* c2)))
        ((inproduct? c1) (append c1 (cons '* (list c2))))
        ((inproduct? c2) (append (list c1) (cons '* c2)))
        (else (list c1 '* c2))))

(define (inmake-sum c1 c2)
  (cond ((eq? c1 0) c2)
        ((eq? c2 0) c1)
        ((and (number? c1) (number? c2)) (+ c1 c2))
        ((and (insum? c1) (insum? c2)) (append c1 (cons '+ c2)))
        ((insum? c1) (append c1 (cons '+ (list c2))))
        ((insum? c2) (append (list c1) (cons '+ c2)))
        (else (list c1 '+ c2))))

(define addend inaddend)
(define augend inaugend)
(define multiplier inmultiplier)
(define multiplicand inmultiplicand)
(define sum? insum?)
(define product? inproduct?)
(define make-sum inmake-sum)
(define make-product inmake-product)

(define (make-set . l) l)
(define (element-of s1 e)
  (cond ((null? s1) #f)
        ((equal? e (car s1)) #t)
        (else (element-of (cdr s1) e))))
(define (difset s1 s2)
  (if (null? s1)
      s1
      (append 
       (if (element-of s2 (car s1))
           '()
           (list (car s1)))
       (difset (cdr s1) s2))))
(define (directsum s1 s2)
  (append s1 s2))
(define (union-set s1 s2)
  (directsum (difset s1 s2) s2))
(define (interset s1 s2)
  (cond ((or (null? s1) (null? s2)) '())
        ((element-of s2 (car s1))
         (cons (car s1)
               (interset (cdr s1) s2)))
        (else (interset (cdr s1) s2))))
(define (interset2 s1 s2)
  (difset s1 (difset s1 s2)))
(define (equalset? s1 s2)
  (cond ((null? s1) #t)
        ((not (element-of s2 (car s1))) #f)
        (else (equalset? (cdr s1) s2))))
(define (adjoin-set x set)
  (if (element-of set x)
      set
      (cons x set)))
(define (element-of-oset set x)
  (cond ((null? set) #f)
        ((= x (car set)) #t)
        ((> x (car set)) #f)
        (else (element-of-oset (cdr set) x))))
(define (adjoin-oset set x)
  (cond ((null? set) (list x))
        ((< x (car set)) (cons x set))
        ((= x (car set)) set)
        (else (cons (car set) (adjoin-oset (cdr set) x)))))

(define (intersection-oset s1 s2)
  (if (or (null? s1) (null? s2)) null
      (let ((x1 (car s1)) (x2 (car s2)))
        (cond ((= x1 x2)
               (cons x1 (intersection-oset (cdr s1) (cdr s2))))
              ((< x1 x2) 
               (intersection-oset (cdr s1) s2))
              ((> x1 x2) 
               (intersection-oset s1 (cdr s2)))))))

(define (union-oset s1 s2)
  (if (or (null? s1) (null? s2)) (append s1 s2)
      (let ((x1 (car s1)) (x2 (car s2)))
        (cond ((= x1 x2)
               (cons x1 (union-oset (cdr s1) (cdr s2))))
              ((< x1 x2) 
               (cons x1 (union-oset (cdr s1) s2)))
              ((> x1 x2) 
               (cons x2 (union-oset s1 (cdr s2))))))))

(define (key tree) (car tree))
(define (left tree) (cadr tree))
(define (right tree) (caddr tree))
(define (make-tree key left right) (list key left right))
(define (find x tree)
  (cond ((null? set) #f)
        ((= x (key tree)) #t)
        ((< x (key tree)) (find x (left tree)))
        ((> x (key tree)) (find x (right tree)))))
(define (insert x tree)
  (cond ((null? tree) (make-tree x null null))
        ((= x (key tree)) tree)
        ((< x (key tree)) (make-tree (key tree) (insert x (left tree)) (right tree)))
        ((> x (key tree)) (make-tree (key tree) (left tree) (insert x (right tree))))))
(define (listtree tree)
  (if (null? tree) 
      null
      (append (listtree (left tree))
              (cons (key tree)
                    (listtree (right tree))))))
(define (listtree2 tree)
  (define (copy tree list)
    (if (null? tree)
        list
        (copy (left tree) 
              (cons (key tree)
                    (copy (right tree)
                          list)))))
  (copy tree '()))
        
(define (treefromlistr l)
  (if (null? l) null
      (insert (car l) (treefromlistr (cdr l)))))
(define (treefromlist l)
  (treefromlistr (reverse l)))

(define (treelist l)
  (car (tl l (length l))))


(define (listit a b)
  (if (> a b) '()
      (cons a (listit (+ 1 a) b))))

(define (tl l n)
  (if (or (null? l) (= n 0)) 
      (cons null l)
      (let ((ls (quotient (- n 1) 2)))
        (let ((lefttree1 (tl l ls)))
          (let ((lefttree (car lefttree1))
                (nonleft (cdr lefttree1))
                (rs (- n (+ ls 1))))
            (let ((thiss (car nonleft))
                  (righttree1 (tl (cdr nonleft) rs)))
              (let ((righttree (car righttree1))
                    (rem (cdr righttree1)))
                (cons (make-tree thiss lefttree righttree) 
                      rem))))))))
      
;(define (shuffle l)
;(define (make-bin-list l)

(define (uniontree t1 t2)
  (treelist (union-oset (listtree t1) (listtree t2))))

(define (intersectiontree t1 t2)
  (treelist (intersection-oset (listtree t1) (listtree t2))))

(define (make-leaf symbol weight)
  (list 'leaf symbol weight))
(define (leaf? object)
  (eq? (car object) 'leaf))
(define (symbol-leaf x) (cadr x))
(define (weight-leaf x) (caddr x))

(define (make-code-tree left right)
  (list left right (append (symbols left) (symbols right))
        (+ (weight left) (weight right))))
(define (left-branch tree) (car tree))
(define (right-branch tree) (cadr tree))
(define (symbols tree)
  (if (leaf? tree)
      (list (symbol-leaf tree))
      (caddr tree)))
(define (weight tree)
  (if (leaf? tree)
      (weight-leaf tree)
      (cadddr tree)))

(define (decode bits tree)
  (define (decode-1 bits current-branch)
    (if (null? bits) null
        (let ((next-branch (choose-branch (car bits) current-branch)))
          (if (leaf? next-branch)
              (cons (symbol-leaf next-branch)
                    (decode-1 (cdr bits) tree))
              (decode-1 (cdr bits) next-branch)))))
  (decode-1 bits tree))

(define (choose-branch bit branch)
  (cond ((= bit 0) (left-branch branch))
        ((= bit 1) (right-branch branch))
        (else (error "lol"))))

(define (adjoinh x set)
  (cond ((null? set) (list x))
        ((< (weight x) (weight (car set))) (cons x set))
        (else (cons (car set) (adjoinh x (cdr set))))))

(define (make-leaf-set lp)
  (if (null? lp) null
      (let ((pair (car lp)))
        (adjoinh (make-leaf (car pair) (cadr pair))
                 (make-leaf-set (cdr lp))))))
                 
(define (encode message tree)
  (if (null? message)
      '()
      (append (encode-symbol (car message) tree)
              (encode (cdr message) tree))))

(define (encode-symbol symb tree)
  (cond ((not (memq symb (symbols tree))) (error "not available"))
        ((leaf? tree) '())
        ((memq symb (symbols (left-branch tree))) (cons 0 (encode-symbol symb (left-branch tree))))
        (else (cons 1 (encode-symbol symb (right-branch tree))))))

(define (generate-huffman-tree pairs)
  (successive-merge (make-leaf-set pairs)))

(define (g p) (generate-huffman-tree p))

(define (successive-merge l)
  (suc (maketreelist l)))

(define (maketreelist l)
  (if (null? l) l
      (cons (cons (cons (symbols (car l)) (weight (car l)))
                  (car l))
            (maketreelist (cdr l)))))

(define (suc l)
  (if (= (length l) 1) (cdar l)
      (suc (adjoinhuf (cddr l)
                      (cons (cons (append (caaar l) (caaadr l)) (+ (cdaar l) (cdaadr l)))
                            (make-code-tree (cdar l) (cdadr l)))))))

(define (adjoinhuf lst el)
  (cond ((null? lst) (list el))
        ((< (cdar el) (cdaar lst)) (cons el lst))
        (else (cons (car lst) (adjoinhuf (cdr lst) el)))))

(define message '(a a a c d f d a s d c s d a s s s d d d d f f c c d a s s a d f f c x d a c a d f j s d a))

(define (insert pairs letter)
  (cond ((null? pairs) (list (list letter 1)))
        ((eq? letter (caar pairs)) (cons (list letter (+ 1 (cadar pairs))) (cdr pairs)))
        (else (cons (car pairs) (insert (cdr pairs) letter)))))

(define (genfreq mes)
  (if (null? mes) '()
      (insert (genfreq (cdr mes)) (car mes))))

(define code (encode message (g (genfreq message))))

(define dcd (decode code (g (genfreq message))))

(equal? dcd message)
