(define setcar set-car!)
(define setcdr set-cdr!)


(define (append x y)
  (if (null? x)
      y
      (cons (car x) (append (cdr x) y))))

(define (append! x y)
  (setcdr (lastpair x) y)
  x)

(define (lastpair x)
  (if (null? (cdr x))
      x
      (lastpair (cdr x))))

(define l1 '(a b c d e f g h i j))
(define l2 '(k l m n o p q))

(append l1 l2)
(append! l1 l2)

(define (makecyc l)
  (setcdr (lastpair l) l))

(define z (makecyc (list 'a 'b 'c)))

(define (myst a)
  (define (loop x y)
    (if (null? x) y
        (let ((t (cdr x)))
          (setcdr x y)
          (loop t x))))
  (loop a '()))

(define (countp x)
  (define seen '())
  (define (iter x)
    (if (not (pair? x)) 0
        (+ (iter (car x))
           (iter (cdr x))
           (if (memq x seen) 0
               (begin
                 (set! seen (cons x seen))
                 1)))))
  (iter x))

(define l3 (cons l1 l2))

(countp l1)

(define lop '(a b c d e f g))
(set-cdr! (cdddr (cdddr lop)) (cddr lop))

(define (loop x)
  (define seen '())
  (define (iter x)
    (cond ((null? x) #f)
          ((memq x seen) #t)
          (else 
           (set! seen (cons x seen))
           (iter (cdr x)))))
  (iter x))

(loop l1)
(loop lop)

(define (constloop x)
  (define (find x k p y c)
    (if (null? x)
        'noloop
        (if (= 0 k) 
            (if (and (eq? x y) c)
                'loop
                (if (= 0 p) 
                    'unknown
                    (find (cdr x) 0 (- p 1) y #t)))
            (find (cdr x) (- k 1) p (cdr x) #f))))
  (define (iter l)
    (if (eq? (find x l l x #f) 'unknown) 
        (iter (* l 2))
        (find x l l x #f)))
  (iter 1))

(constloop l1)
(constloop lop)

;for list of n elements loop and constloop have the following properties:
;constloop - runtime O(n*logn), space O(1) - using idea similar to binary search
;loop - runtime O(n^2), space O(n) - can be made with runtime O(n) using hash table
;both use constant space for recursion because they are tail recursive
;space requirement only measures the additional required space (still needed O(n) for the given list)

(define (mq) (cons '() '()))
(define (frontend q) (car q))
(define (rearend q) (cdr q))
(define (print q) (frontend q))
(define (setfront q a) (set-car! q a))
(define (setrear q a) (set-cdr! q a))
(define (empty? q) (null? (frontend q)))
(define (push q a)
  (if (empty? q) 
      (begin (setfront q (list a)) (setrear q (frontend q)))
      (begin (set-cdr! (rearend q) (list a)) (setrear q (cdr (rearend q)))))
  (print q))
(define (pop q) 
  (begin (if (empty? q) #f (setfront q (cdr (frontend q))))
         (if (empty? q) (clear q))))
(define (front q) (if (empty? q) #f (car (frontend q))))
(define (delete q) (pop q) (print q))
(define (clear q) (begin (setfront q '()) (setrear q '())))

(define q (mq))
(push q 'a)
(push q 'b)
(delete q)
(push q 'c)
(push q 'd)
(delete q)

(define (queue)
  (define frontend '())
  (define rearend '())
  (define q (cons frontend rearend))
  (define (dispatch m)
    (cond ((eq? m 'push) (lambda (x) (push q x)))
          ((eq? m 'pop) (lambda () (pop q)))
          ((eq? m 'empty?) (lambda () (empty? q)))
          ((eq? m 'front) (lambda () (front q)))
          ((eq? m 'clear) (lambda () (clear q)))
          ((eq? m 'print) (lambda () (print q)))
          ((eq? m 'delete) (lambda () (delete q)))))
  dispatch)

(define q (queue))
((q 'push) 1)
((q 'push) 2)
((q 'pop))
((q 'delete))

(define (dequeue)
  (define frontend null)
  (define rearend null)
  (define (data e) (car e))
  (define (next e) (caddr e))
  (define (prev e) (cadr e))
  (define (setdata e a) (set-car! e a))
  (define (setnext e a) (set-car! (cddr e) a))
  (define (setprev e a) (set-car! (cdr e) a))
  (define (setrear a) (set! rearend a))
  (define (setfront a) (set! frontend a))
  (define (empty?)  (null? frontend))
  (define (print a)
    (if (null? a) a
        (cons (data a) (print (next a)))))
  (define (clear) (begin (setfront null) (setrear null)))
  (define (push_back a)
    (if (empty?)
        (begin (setfront (list a null null)) (setrear frontend))
        (begin (setnext rearend (list a null null))
               (setprev (next rearend) rearend)
               (setrear (next rearend)))))
  (define (pop_back)
    (cond ((empty?) #f)
          ((eq? frontend rearend) (begin (setfront null) (setrear null)))
          (else (begin (setrear (prev rearend)) (setnext rearend null)))))
  (define (push_front a)
    (if (empty?) (push_back a)
        (begin (setprev frontend (list a null null))
               (setnext (prev frontend) frontend)
               (setfront (prev frontend)))))
  (define (pop_front)
    (cond ((empty?) #f)
          ((eq? frontend rearend) (begin (setfront null) (setrear null)))
          (else (begin (setfront (next frontend)) (setprev frontend null)))))
  (define (front) (data frontend))
  (define (back) (data rearend))
  (define (dispatch . m)
    (if (null? m) (print frontend)
        (cond ((eq? (car m) 'push_back) (lambda (x) (push_back x) (print frontend)))
              ((eq? (car m) 'print) (lambda () (print frontend)))
              ((eq? (car m) 'pop_back) (lambda () (pop_back) (print frontend)))
              ((eq? (car m) 'empty?) empty?)
              ((eq? (car m) 'front) front)
              ((eq? (car m) 'back) back)
              ((eq? (car m) 'clear) (lambda () (clear) (print frontend)))
              ((eq? (car m) 'push_front) (lambda (x) (push_front x) (print frontend)))
              ((eq? (car m) 'pop_front) (lambda () (pop_front) (print frontend))))))
  dispatch)

(define deq (dequeue))
((deq 'push_back) 1)
((deq 'pop_back))
((deq 'pop_front))
((deq 'pop_back))
((deq 'push_front) 2)
((deq 'pop_front))
((deq 'push_back) 1)
((deq 'push_front) 2)
((deq 'push_back) 3)
((deq 'push_front) 4)
((deq 'push_back) 5)
((deq 'push_front) 6)
((deq 'print))
((deq 'empty?))
((deq 'clear))
((deq 'empty?))


(define (lookup1 key table)
  (let ((record (assoc key (cdr table))))
    (if record (cdr record)
        record)))
(define (assoc key l)
  (cond ((null? l) #f)
        ((equal? key (caar l)) (car l))
        (else (assoc key (cdr l)))))
(define (insert1 key value table)
  (let ((record (assoc key (cdr table))))
    (if record (set-cdr! record value)
        (set-cdr! table (cons (cons key value) (cdr table))))))
(define (maketable) (list 'table))

(define t1 (maketable))
(lookup1 1 t1)
(insert1 1 11 t1)
(lookup1 1 t1)
(insert1 2 22 t1)
(insert1 1 10 t1)
(lookup1 2 t1)
(lookup1 1 t1)

(define (lookup key1 key2 t)
  (let ((subtable (assoc key1 (cdr t))))
    (if subtable
        (let ((record (assoc key2 (cdr subtable))))
          (if record
              (cdr record)
              record))
        subtable)))

(define (insert k1 k2 val t)
  (let ((subtable (assoc k1 (cdr t))))
    (if subtable
        (let ((record (assoc k2 (cdr subtable))))
          (if record
              (set-cdr! record val)
              (set-cdr! subtable (cons (cons k2 val) (cdr subtable)))))
        (set-cdr! t (cons (list k1 (cons k2 val))
                          (cdr t))))))


(define (multi . comp)
  (define equal? (if (null? comp) equal? (car comp))) 
  (define t (cons (cons 'table null) null))
  (define (assoc t v)
    (cond ((null? t) #f)
          ((equal? (caaar t) v) (car t))
          (else (assoc (cdr t) v))))
  (define (find t . k)
    (if (null? k) (if (null? (cdar t)) #f t)
        (let ((r (assoc (cdr t) (car k))))
          (if (not r) #f
              (apply find r (cdr k))))))
  (define (look t . k)
    (let ((r (apply find t k)))
      (if r (cadar r) r)))
  (define (insert t v . k)
    ;(display t)
    ;(display "\n")
    (if (null? k) 
        (if (null? (cdar t)) 
            (set-cdr! (car t) (cons v null))
            (set-car! (cdar t) v))
        (let ((r (assoc (cdr t) (car k))))
          (if r 
              (apply insert r v (cdr k))
              (begin (setcdr t (cons (cons (cons (car k) null) null) (cdr t)))
                     (apply insert (cadr t) v (cdr k)))))))
  (define (dispatch m . arg)
    (cond ((eq? m 'look) (apply look t arg))
          ((eq? m 'insert) (apply insert t arg))
          (else #f)))
  dispatch)

(define (look t . k)
  (apply t 'look k))
(define (insert t v . k)
  (apply t 'insert v k))

;(define t (multi (lambda (x y) (= (remainder x 3) (remainder y 3)))))
;(define t (multi))


(define (memoize f)
  (let ((table (multi equal?)))
    (lambda l (let ((prev (apply look table l)))
                (or prev
                    (let ((res (apply f l)))
                      (apply insert table res l)
                      res))))))

(define (fib n) (if (<= n 1) n (+ (fib (- n 1)) (fib (- n 2)))))

(define fi (lambda () 1))
(set! fi fib)
(set! fib (memoize fib))

(fib 1)
(fib 0)
(fib 1022)
(fi 30)