(define true #t)
(define false #f)
(define sapply apply)

(define evaltable '())
(define (puteval a c)
  (set! evaltable (cons (cons a c) evaltable)))
(define (geteval a)
  (define (go l)
    (cond ((null? l) #f)
          ((equal? a (caar l)) (cdar l))
          (else (go (cdr l)))))
  (go evaltable))

(define (getfun exp)
  (cond ((variable? exp) findvarval)
        ((selfeval? exp) evalselfeval)
        (else (let ((r (geteval (car exp))))
                (or r
                    (cond ((application? exp) evalapply)
                          (else (error "undetermined type\n"))))))))

(define (eval exp env)
  (let ((f (getfun exp)))
    (f exp env)))

(define (eval2 exp env)
  (cond ((selfeval? exp) (evalselfeval exp env))
        ((variable? exp) (findvarval exp env))
        ((quoted? exp) (evalquote exp env))
        ((assignment? exp) (evalassign exp env))
        ((definition? exp) (evaldef exp env))
        ((undef? exp) (evalundef exp env))
        ((or? exp) (evalor exp env))
        ((time? exp) (evaltime exp env))
        ((and? exp) (evaland exp env))
        ((if? exp) (evalif exp env))
        ((cond? exp) (evalcond exp env))
        ((let? exp) (evallet exp env))
        ((let*? exp) (evallet* exp env))
        ((lambda? exp) (evallambda exp env))
        ((begin? exp) (evalbegin exp env))
        ((application? exp) (evalapply exp env))
        (else (error "eval"))))

(define (evalselfeval exp env) exp)
(define (evalapply exp env) 
  (apply (eval (operator exp) env)
         (listofvals (operands exp) env)))

(define (apply proc args)
  (cond ((primitiveproc? proc)
         (applyprimitive proc args))
        ((compoundproc? proc)
         (evalseq 
          (procbody proc)
          (extendenv
           (procparams proc)
           args
           (procenv proc))))
        (else "apply")))

(define (install-begin)
  (define (evalbegin exp env) (evalseq (beginactions exp) env))
  (puteval 'begin evalbegin)
  'begin-installed)

(define (makecall proc args)
  (cons proc args))

(define (listofvalsleftright exps env)
  (define now 1)(define then 1)
  (if (noops? exps)
      null
      (begin (set! now (eval (firstoperand exps) env))
             (set! then (listofvals (restoperands exps) env))
             (cons now then))))

(define (listofvalsrightleft exps env)
  (define now 1)(define then 1)
  (if (noops? exps)
      null
      (begin (set! then (listofvals (restoperands exps) env))
             (set! now (eval (firstoperand exps) env))
             (cons now then))))

(define (listofvals exps env) (listofvalsleftright exps env))

(define (install-if)
  (define (if? exp) (taggedlist? exp 'if))
  (define (ifpredicate exp) (cadr exp))
  (define (ifconsequent exp) (caddr exp))
  (define (ifalternative exp)
    (if (not (null? (cdddr exp)))
        (cadddr exp)
        '#f))
  (define (evalif exp env)
    (if (true? (eval (ifpredicate exp) env))
        (eval (ifconsequent exp) env)
        (eval (ifalternative exp) env)))
  (puteval 'if evalif)
  'if-installed)

(define (evalseq seq env)
  (cond ((lastexp? seq) (eval (firstexp seq) env))
        (else (eval (firstexp seq) env)
              (evalseq (restexps seq) env))))
(define (makeseq . l)
  (cons 'begin l))

(define (install-assign)
  (define (setvarval var val env)
    (define (envloop env)
      (define (scan vars vals)
        (cond ((null? vars) (envloop (encloseenv env)))
              ((eq? var (car vars)) (set-car! vals val))
              (else (scan (cdr vars) (cdr vals)))))
      (if (eq? env emptyenv)
          (error "unbound variable to set" var)
          (let ((frame (firstframe env)))
            (scan (framevarnames frame) 
                  (framevarvals frame)))))
    (envloop env))
  (define (evalassign exp env)
    (setvarval (assignmentvar exp)
               (eval (assignmentval exp) env)
               env))
  (puteval 'set! evalassign)
  'set!-installed)

(define (selfeval? exp)
  (cond ((number? exp) true)
        ((string? exp) true)
        ((boolean? exp) true)
        (else false)))

(define (variable? exp) (symbol? exp))

(define (quoted? exp)
  (taggedlist? exp 'quote))


(define (install-quote)
  (define (textofquote exp) (cadr exp))
  (define (evalquote exp env) (textofquote exp))
  (puteval 'quote evalquote)
  'quote-installed)

(define (taggedlist? exp tag)
  (if (pair? exp)
      (eq? (car exp) tag)
      false))

(define (assignment? exp)
  (taggedlist? exp 'set!))
(define (assignmentvar exp) (cadr exp))
(define (assignmentval exp) (caddr exp))

(define (defvar! var val env)
    (let ((frame (firstframe env)))
      (define (scan vars vals)
        (cond ((null? vars) (bindtoframe var val frame))
              ((eq? var (car vars)) (set-car! vals val))
              (else (scan (cdr vars) (cdr vals)))))
      (scan (framevarnames frame)
            (framevarvals frame))))
(define (install-define)
  (define (1undefvar! var env)
    (define (envloop env)
      (define (scan vars vals)
        (cond ((null? vars) (error "variable not found"))
              ((eq? var (car vars)) false)
              (else (if (not (scan (cdr vars) (cdr vals)))
                        (begin (set-cdr! vals (cddr vals))
                               (set-cdr! vars (cddr vars)))
                        true))))
      (if (eq? env emptyenv)
          (error "unbound variable" var)
          (let ((frame (firstframe env)))
            (if (not (scan (framevarnames frame)
                           (framevarvals frame)))
                (set! frame (null null))))))
    (envloop env))
  (define (undefvar! var env) (1undefvar! var env))
  (define (defvar exp)
    (if (symbol? (cadr exp))
        (cadr exp)
        (caadr exp)))
  (define (defval exp)
    (display exp)
    (newline)
    (if (symbol? (cadr exp))
        (caddr exp)
        (makelambda (cdadr exp)
                    (caddr exp))))
  (define (undefvarvar exp) (cadr exp))
  (define (evalundef exp env)
    (undefvar! (undefvarvar exp) env))
  (define (evaldef exp env)
    (defvar! 
      (defvar exp)
      (eval (defval exp) env)
      env))
  (puteval 'undef evalundef)
  (puteval 'define evaldef)
  'define/undefine-installed)

(define (undef? exp) (taggedlist? exp 'undef))
(define (undefvarvar exp) (cadr exp))
(define (definition? exp)
  (taggedlist? exp 'define))
(define (defvar exp)
  (if (symbol? (cadr exp))
      (cadr exp)
      (caadr exp)))
(define (defval exp)
  (display exp)
  (newline)
  (if (symbol? (cadr exp))
      (caddr exp)
      (makelambda (cdadr exp)
                   (caddr exp))))

(define (makedefvar var val) (list 'define var val))
(define (makedefproc name args body) (list 'define (cons name args) body))

(define (install-lambda)
  (define (lambdaparams exp) (cadr exp))
  (define (lambdabody exp) (cddr exp))
  (define (evallambda exp env) 
    (makeproc (lambdaparams exp)
              (lambdabody exp)
              env))
  (puteval 'lambda evallambda)
  'lambda-installed)

(define (lambda? exp) (taggedlist? exp 'lambda))

(define (makelambda params body)
  (list 'lambda params body))

(define (makeif predicate consequent alternative)
  (list 'if predicate consequent alternative))

(define (begin? exp) (taggedlist? exp 'begin))
(define (beginactions exp) (cdr exp))
(define (lastexp? seq) (null? (cdr seq)))
(define (firstexp seq) (car seq))
(define (restexps seq) (cdr seq))

(define (sequencetoexp seq)
  (cond ((null? seq) seq)
        ((lastexp? seq) (firstexp seq))
        (else (makebegin seq))))
(define (makebegin seq) (cons 'begin seq))


(define (1application? exp) (pair? exp))
(define (1operator exp) (car exp))
(define (1operands exp) (cdr exp))
(define (42application? exp) (taggedlist? exp 'call))
(define (42operator exp) (cadr exp))
(define (42operands exp) (cddr exp))

(define (application? exp) (1application? exp))
(define (operator exp) (1operator exp))
(define (operands exp) (1operands exp))

(define (noops? ops) (nooperands? ops))
(define (nooperands? ops) (null? ops))
(define (firstoperand ops) (car ops))
(define (restoperands ops) (cdr ops))

(define (install-cond)
  (define (cond? exp) (taggedlist? exp 'cond))
  (define (condclauses exp) (cdr exp))
  (define (condpredicate clause) (car clause))
  (define (condactions clause) (cdr clause))
  (define (condtoif exp) (expandclauses (condclauses exp)))
  (define (condelseclause? clause) (eq? (condpredicate clause) 'else))
  (define (condspecialclause? clause) 
    (and (not (null? (cdr clause))) (eq? (cadr clause) '=>)))
  (define (condnormalclause? clause)
    (and (not (condelseclause? clause)) (not (condspecialclause? clause))))
  (define (expandclauses clauses)
    (if (null? clauses) 
        '#f
        (let ((first (car clauses))
              (rest (cdr clauses)))
          (if (condelseclause? first)
              (if (null? rest)
                  (sequencetoexp (condactions first))
                  (error "expand clause" clauses))
              (makeif (condpredicate first)
                      (sequencetoexp (condactions first))
                      (expandclauses rest))))))
  
  (define (evalcondclauses clauses env)
    (if (null? clauses) '#f
        (let ((first (car clauses))
              (rest (cdr clauses)))
          (cond ((condelseclause? first)
                 (if (null? rest)
                     (eval (sequencetoexp (condactions first)) env)
                     (error "evalcondclauses" clauses)))
                ((condnormalclause? first)
                 (if (eval (condpredicate first) env)
                     (eval (sequencetoexp (condactions first)) env)
                     (evalcondclauses rest env)))
                ((condspecialclause? first)
                 (let ((pred (condspecialpredicate first)))
                   (if pred
                       (apply (eval (condspecualproc first) env) pred)
                       (evalcondclauses rest env))))
                (else "error - unrecognized clause" first)))))
  
  
  (define (1evalcond exp env) (eval (condtoif exp) env))
  (define (2evalcond exp env)
    (evalcondclauses (condclauses exp) env))
  
  (define (evalcond exp env) (2evalcond exp env))
  
  (puteval 'cond evalcond)
  'cond-installed)

(define (install-or)
  (define (expandor clauses)
    (if (null? clauses) '#f
        (makeif (car clauses) (car clauses) (expandor (cdr clauses)))))
  (define (orclauses exp) (cdr exp))
  (define (2evalor exp env)
    (eval (expandor (orclauses exp)) env))
  (define (orevalclauses clauses env)
    (if (null? clauses) '#f
        (let ((r (eval (car clauses) env)))
          (if r r (orevalclauses (cdr clauses) env)))))            
  (define (1evalor exp env)
    (orevalclauses (orclauses exp) env))
  (define (evalor exp env) (1evalor exp env))
  (puteval 'or evalor)
  'or-installed)

(define (install-and)
  (define (expandand clauses)
    (if (null? clauses) '#t
        (makeif (car clauses) (expandand (cdr clauses)) (car clauses))))
  (define (andclauses exp) (cdr exp))
  (define (2evaland exp env)
    (eval (expandand (andclauses exp)) env))
  (define (andevalclauses clauses env)
    (if (null? clauses) '#t
        (let ((r (eval (car clauses) env)))
          (if r (andevalclauses (cdr clauses) env) r))))            
  (define (1evaland exp env)
    (andevalclauses (andclauses exp) env))
  (define (evaland exp env) (1evaland exp env))
  (puteval 'and evaland)
  'and-installed)

(define (install-let)
  (define (letnormal? exp) (null? (cdddr exp)))
  (define (letspecial? exp) (not (letnormal? exp)))
  (define (letdeclarations exp) (cadr exp))
  (define (letbody exp) (caddr exp))
  (define (letspecialdecl exp) (caddr exp))
  (define (letspecialproc exp) (cadr exp))
  (define (letspecialbody exp) (cadddr exp))
  (define (letdeclvar decl) (car decl))
  (define (letdeclval decl) (cadr decl))
  (define (lettolambda exp)
    (let ((body (letbody exp))
          (decl (letdeclarations exp)))
      (cons (makelambda (map letdeclvar decl) body)
            (map letdeclval decl))))
  (define (letspecialtolambda exp)
    (let ((proc (letspecialproc exp))
          (body (letspecialbody exp))
          (decl (letspecialdecl exp)))
      (makeseq (makedefproc proc (map letdeclvar decl) body)
               (makecall proc (map letdeclval decl)))))
  (define (evallet exp env) 
    (if (letspecial? exp)
        (eval (letspecialtolambda exp) env)
        (eval (lettolambda exp) env)))  
  (define (nestlet decl body)
    (if (null? decl) (list 'let '() body)
        (makelet (list (car decl)) (nestlet (cdr decl) body))))
  (define (let*? exp) (taggedlist? exp 'let*))
  (define (evallet* exp env)
    (eval (nestlet (letdeclarations exp) (letbody exp)) env))
  
  (puteval 'let* evallet*)
  (puteval 'let evallet)
  'let-installed)

(define (makelet decl body)
  (list 'let decl body))

;(define exp '(let* ((x 3) (y (+ x 2)) (z (+ x y 5))) (* x y z)))
;(nestlet (letdeclarations exp) (letbody exp))

(define (install-loops)
  (define (while? exp) (taggedlist? exp 'while))
  (define (whilepred exp) (cadr exp))
  (define (whileexp exp) (caddr exp))
  (define (evalwhile exp env)
    (let ((r (eval (whilepred exp) env)))
      (if r
          (begin (eval (whileexp exp) env)
                 (eval exp env))
          r)))  
  (define (for? exp) (taggedlist? exp 'for))
  (define (forpred exp) (caddr exp))
  (define (forexp exp) (caddddr exp))
  (define (forafter exp) (cadddr exp))
  (define (forbegin exp) (cadr exp))
  (define (evalforiter exp env)
    (let ((r (eval (forpred exp) env)))
      (if r 
          (begin (eval (forexp exp) env)
                 (eval (forafter exp) env)
                 (evalforiter exp env)))))
  (define (evalfor exp env)
    (begin (eval (forbegin exp) env)
           (evalforiter exp env)))
  
  (puteval 'for evalfor)
  (puteval 'while evalwhile)
  'loops-installed)


(define (install-time)
  (define (time? exp) (taggedlist? exp 'time))
  (define (timeexp exp) (cadr exp))
  (define (evaltime exp env)
    (time (eval (timeexp exp) env)))
  (puteval 'time evaltime)
  'time-installed)

(define (true? x) (not (false? x)))
(define (false? x) (eq? x false))

(define (applyprimitive proc args) (sapply (primitiveimpl proc) args))
(define (primitiveproc? proc) (primitive? proc))
(define (makeproc params body env)
  (display body)
  (newline)
  (list 'procedure params body env))
(define (compoundproc? proc) (taggedlist? proc 'procedure))
(define (procparams proc) (cadr proc))
(define (procbody proc) (caddr proc))
(define (procenv proc) (cadddr proc))


(define (encloseenv env) (cdr env))
(define (firstframe env) (car env))
(define emptyenv '())

(define (makeframe vars vals)
  (cons vars vals))
(define (framevarnames frame) (car frame))
(define (framevarvals frame) (cdr frame))
(define (bindtoframe var val frame)
  (set-car! frame (cons var (car frame)))
  (set-cdr! frame (cons val (cdr frame))))

(define (extendenv vars vals env)
  (if (= (length vals) (length vars))
      (cons (makeframe vars vals) env)
      (if (< (length vals) (length vars))
          (error "too many args" vars vals)
          (error "too few  args" vars vals))))

(define (findvarval var env)
  (define (envloop env)
    (define (scan vars vals)
      (cond ((null? vars) (envloop (encloseenv env)))
            ((eq? var (car vars)) (car vals))
            (else (scan (cdr vars) (cdr vals)))))
    (if (eq? env emptyenv)
        (error "unbound variable" var)
        (let ((frame (firstframe env)))
          (scan (framevarnames frame) 
                (framevarvals frame)))))
  (envloop env))
            
(define (primitiveproc? proc)
  (taggedlist? proc 'primitive))

(define (primitiveimpl proc) (cadr proc))

(define primelist  '(car cdr cons - + * / remainder null? <= < = > >= equal? eq? set-cdr! set-car! map  apply))
(define primes (list car cdr cons - + * / remainder null? <= < = > >= equal? eq? set-cdr! set-car! map sapply))

(define (getprimitives names objs)
  (if (or (null? names) (null? objs))
      null
      (cons (list (car names) (car objs))
            (getprimitives (cdr names) (cdr objs)))))

(define primef (getprimitives primelist primes))

(define (primitivenames) (map car primef))
(define (primitives) (map (lambda (p) (list 'primitive (cadr p))) primef))


(define (setupenv)
  (let ((initial (extendenv (primitivenames)
                            (primitives)
                            emptyenv)))
    (defvar! '#t #t initial)
    (defvar! '#f #f initial)
    (defvar! 'null null initial)
    initial))

(define globalenv (setupenv))

(define inputprompt "----- eval input ------\n")
(define outputprompt "----- eval output -----\n")

(define (driver)
  (promptforinput inputprompt)
  (let ((input (read)))
    (let ((output (eval input globalenv)))
      (announceoutput outputprompt)
      (userprint output)))
  (driver))

(define (promptforinput string) (display string))
(define (announceoutput string) (display string))

(define (userprint obj)
  (if (compoundproc? obj)
      (begin 
        (display (list 'compoundproc (procparams obj) (procbody obj) 'environment))
        (newline))
      (begin
        (display obj)
        (newline))))

(define (fib n) (if (<= n 1) n (+ (fib (- n 1)) (fib (- n 2)))))

(install-if)
(install-begin)
(install-lambda)
(install-define)
(install-quote)
(install-cond)
(install-let)
(install-time)
(install-or)
(install-and)
(install-assign)
(install-loops)

(eval '(define (fib n) (if (<= n 1) n (+ (fib (- n 1)) (fib (- n 2))))) globalenv)

(driver)